/*
 * Copyright 2012 Calxeda, Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <linux/kernel.h>
#include <linux/gfp.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/err.h>
#include <linux/io.h>
#include <linux/spinlock.h>
#include <linux/device.h>
#include <linux/of_device.h>
#include <linux/of_address.h>
#include <linux/platform_device.h>
#include <linux/ahci_platform.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/export.h>
#include "../../../drivers/ata/ahci.h"

#define CPHY_MAP(dev, addr) ((((dev) & 0x1f) << 7) | (((addr) >> 9) & 0x7f))
#define CPHY_ADDR(addr) (((addr) & 0x1ff) << 2)
#define SERDES_CR_CTL			0x80a0
#define SERDES_CR_ADDR			0x80a1
#define SERDES_CR_DATA			0x80a2
#define CR_BUSY				0x0001
#define CR_START			0x0001
#define CR_WR_RDN			0x0002
#define CPHY_RX_INPUT_STS		0x2002
#define CPHY_SATA_OVERRIDE	 	0x4000
#define CPHY_OVERRIDE			0x2005
#define SPHY_LANE			0x100
#define SPHY_HALF_RATE			0x0001
#define CPHY_SATA_DPLL_MODE		0x0700
#define CPHY_SATA_DPLL_SHIFT		8
#define CPHY_SATA_DPLL_RESET		(1 << 11)
#define CPHY_PHY_COUNT			6
#define CPHY_LANE_COUNT			4
#define CPHY_PORT_COUNT			(CPHY_PHY_COUNT * CPHY_LANE_COUNT)

static DEFINE_SPINLOCK(cphy_lock);
static void __iomem *phy_base[CPHY_PORT_COUNT];
/* Each of the 6 phys can have up to 4 sata ports attached to i. Map 0-based
 * sata ports to their phys and then to their lanes within the phys
 */
static u8 lane_mapping[CPHY_PORT_COUNT];
static u8 phy_devs[CPHY_PORT_COUNT];

static u32 __combo_phy_reg_read(u8 sata_port, u32 addr)
{
	u32 data;
	u8 dev = phy_devs[sata_port];
	spin_lock(&cphy_lock);
	writel(CPHY_MAP(dev, addr), phy_base[sata_port] + 0x800);
	data = readl(phy_base[sata_port] + CPHY_ADDR(addr));
	spin_unlock(&cphy_lock);
	return data;
}

static void __combo_phy_reg_write(u8 sata_port, u32 addr, u32 data)
{
	u8 dev = phy_devs[sata_port];
	spin_lock(&cphy_lock);
	writel(CPHY_MAP(dev, addr), phy_base[sata_port] + 0x800);
	writel(data, phy_base[sata_port] + CPHY_ADDR(addr));
	spin_unlock(&cphy_lock);
}

static u32 combo_phy_read(u8 sata_port, u32 addr)
{
	while (__combo_phy_reg_read(sata_port, SERDES_CR_CTL) & CR_BUSY)
		udelay(5);
	__combo_phy_reg_write(sata_port, SERDES_CR_ADDR, addr);
	udelay(25);
	__combo_phy_reg_write(sata_port, SERDES_CR_CTL, CR_START);
	udelay(25);
	while (__combo_phy_reg_read(sata_port, SERDES_CR_CTL) & CR_BUSY)
		udelay(5);
	return __combo_phy_reg_read(sata_port, SERDES_CR_DATA);
}

static void combo_phy_write(u8 sata_port, u32 addr, u32 data)
{
	while (__combo_phy_reg_read(sata_port, SERDES_CR_CTL) & CR_BUSY)
		udelay(5);
	__combo_phy_reg_write(sata_port, SERDES_CR_ADDR, addr);
	udelay(25);
	__combo_phy_reg_write(sata_port, SERDES_CR_DATA, data);
	udelay(25);
	__combo_phy_reg_write(sata_port, SERDES_CR_CTL, CR_WR_RDN | CR_START);
}

static void calxeda_cphy_disable_overrides(u8 sata_port)
{
	u8 lane = lane_mapping[sata_port];
	u32 tmp;
	if (unlikely(phy_base[sata_port] == NULL))
		return;
	tmp = combo_phy_read(sata_port, CPHY_RX_INPUT_STS + lane * SPHY_LANE);
	tmp &= ~CPHY_SATA_OVERRIDE;
	combo_phy_write(sata_port, CPHY_OVERRIDE + lane * SPHY_LANE, tmp);
}

static void cphy_override_rx_mode(u8 sata_port, u32 val)
{
	u8 lane = lane_mapping[sata_port];
	u32 tmp;
	tmp = combo_phy_read(sata_port, CPHY_RX_INPUT_STS + lane * SPHY_LANE);
	tmp &= ~CPHY_SATA_OVERRIDE;
	combo_phy_write(sata_port, CPHY_OVERRIDE + lane * SPHY_LANE, tmp);

	tmp |= CPHY_SATA_OVERRIDE;
	combo_phy_write(sata_port, CPHY_OVERRIDE + lane * SPHY_LANE, tmp);

	tmp &= ~CPHY_SATA_DPLL_MODE;
	tmp |= val << CPHY_SATA_DPLL_SHIFT;
	combo_phy_write(sata_port, CPHY_OVERRIDE + lane * SPHY_LANE, tmp);

	tmp |= CPHY_SATA_DPLL_RESET;
	combo_phy_write(sata_port, CPHY_OVERRIDE + lane * SPHY_LANE, tmp);

	tmp &= ~CPHY_SATA_DPLL_RESET;
	combo_phy_write(sata_port, CPHY_OVERRIDE + lane * SPHY_LANE, tmp);

	mdelay(15);
}

static void calxeda_cphy_override_lane(u8 sata_port)
{
	u8 lane = lane_mapping[sata_port];
	u32 tmp, k = 0;
	if (unlikely(phy_base[sata_port] == NULL))
		return;
	do {
		tmp = combo_phy_read(sata_port, CPHY_RX_INPUT_STS +
						lane * SPHY_LANE);
	} while ((tmp & SPHY_HALF_RATE) && (k++ < 1000));
	cphy_override_rx_mode(sata_port, 3);
}

int ahci_calxeda_hardreset(struct ata_link *link, unsigned int *class,
				unsigned long deadline)
{
	const unsigned long *timing = sata_ehc_deb_timing(&link->eh_context);
	struct ata_port *ap = link->ap;
	struct ahci_port_priv *pp = ap->private_data;
	u8 *d2h_fis = pp->rx_fis + RX_FIS_D2H_REG;
	struct ata_taskfile tf;
	bool online;
	u32 sstatus;
	int rc;
	int retry = 10;

	ahci_stop_engine(ap);

	/* clear D2H reception area to properly wait for D2H FIS */
	ata_tf_init(link->device, &tf);
	tf.command = 0x80;
	ata_tf_to_fis(&tf, 0, 0, d2h_fis);

	do {
		calxeda_cphy_disable_overrides(link->ap->port_no);
		rc = sata_link_hardreset(link, timing, deadline, &online, NULL);
		calxeda_cphy_override_lane(link->ap->port_no);

		/* If the status is 1, we are connected, but the link did not
		 * come up. So retry resetting the link again.
		 */
		if (sata_scr_read(link, SCR_STATUS, &sstatus))
			break;
		if (!(sstatus & 0x3))
			break;
	} while (!online && retry--);

	ahci_start_engine(ap);

	if (online)
		*class = ahci_dev_classify(ap);

	return rc;
}
EXPORT_SYMBOL(ahci_calxeda_hardreset);

void __init highbank_register_sata(void)
{
	struct device_node *sata_node;
	int phy_count = 0, phy, port = 0;
	void __iomem *cphy_base[CPHY_PHY_COUNT];
	struct device_node *phy_nodes[CPHY_PHY_COUNT];
	memset(phy_base, 0, sizeof(void *) * CPHY_PORT_COUNT);
	memset(phy_nodes, 0, sizeof(struct device_node*) * CPHY_PHY_COUNT);

	for_each_compatible_node(sata_node, NULL, "calxeda,hb-ahci") {
		do {
			u32 tmp;
			struct device_node *phy_data;
			u32 *lane_data;
			if (of_parse_phandles_with_args(sata_node,
					"calxeda,port-phys", "#phy-cells",
					port, &phy_data,
					(const void **) &lane_data))
				break;
			for (phy = 0; phy < phy_count; phy++) {
				if (phy_nodes[phy] == phy_data)
					break;
			}
			if (phy_nodes[phy] == NULL) {
				phy_nodes[phy] = phy_data;
				cphy_base[phy] = of_iomap(phy_nodes[phy], 0);
				if (cphy_base[phy] == NULL)
					return;
				phy_count += 1;
			}
			lane_mapping[port] = be32_to_cpu(lane_data[0]);
			of_property_read_u32(phy_nodes[phy], "phydev", &tmp);
			phy_devs[port] = tmp;
			phy_base[port] = cphy_base[phy];
			of_node_put(phy_data);
			port += 1;
		} while (port < CPHY_PORT_COUNT);
	}
}

static const struct of_device_id cx_combophy_of_match[] = {
	{ .compatible = "calxeda,hb-combophy", },
	{},
};
MODULE_DEVICE_TABLE(of, cx_combophy_of_match);
